theory KnightTour
  imports Main
    "~~/src/HOL/Library/Code_Target_Nat"
begin

section \<open>Basics\<close>

text \<open>a pair of coordinates\<close>

type_synonym pos = "nat \<times> nat" 

text \<open>knight moves\<close>
(*
 *  . x . x .
 *  x . . . x
 *  . . o . .
 *  x . . . x
 *  . x . x .
 *)

fun knight_move :: "pos \<Rightarrow> pos \<Rightarrow> bool" where
  "knight_move (x1, y1) (x2, y2) \<longleftrightarrow>
   (x1 = x2 + 2 \<or> x2 = x1 + 2) \<and> (y1 = y2 + 1 \<or> y2 = y1 + 1) \<or>
   (x1 = x2 + 1 \<or> x2 = x1 + 1) \<and> (y1 = y2 + 2 \<or> y2 = y1 + 2)"

lemma knight_move_symmetric [ac_simps]:
  "knight_move p q \<longleftrightarrow> knight_move q p"
by (induct rule: knight_move.induct) auto


section \<open>knight tours\<close>

definition knight_tour :: "nat \<Rightarrow> nat \<Rightarrow> pos list \<Rightarrow> bool" where
  "knight_tour n m ps \<longleftrightarrow>
   (\<forall>(x, y) \<in> set ps. x < n \<and> y < m) \<and>
   distinct ps \<and>
   (\<forall>(p, q) \<in> set (zip ps (tl ps)). knight_move p q)"

value "zip [0..<5] (tl [0..<5])"

lemma empty_tour [intro]:
  "knight_tour n m []"
by (auto simp: knight_tour_def)

lemma extend_knight_tour [simp]:
  "knight_tour n m ((x, y) # ps) \<longleftrightarrow>
   knight_tour n m ps \<and>
   x < n \<and> y < m \<and>
   (x, y) \<notin> set ps \<and>
   (ps = [] \<or> knight_move (x, y) (hd ps))"
by (cases ps) (auto simp: knight_tour_def)


text \<open>Compute neighbours.\<close>

definition neighbours_aux :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat list" where
  "neighbours_aux n x d = (if x \<ge> d then [x - d] else []) @ (if x < n - d then [x + d] else [])"

lemma set_neighbours_aux [simp]:
  assumes "x < n"
  shows "set (neighbours_aux n x d) = {x'. x' < n \<and> (x = x' + d \<or> x' = x + d)}"
using assms
by (auto simp only: neighbours_aux_def split: if_splits) auto

fun neighbours :: "nat \<Rightarrow> nat \<Rightarrow> pos \<Rightarrow> pos list" where
  "neighbours n m (x, y) =
   List.product (neighbours_aux n x 1) (neighbours_aux m y 2) @
   List.product (neighbours_aux n x 2) (neighbours_aux m y 1)"

lemma set_neighbours [simp]:
  assumes "x < n" and "y < m"
  shows "set (neighbours n m (x, y)) = {(x', y'). x' < n \<and> y' < m \<and> knight_move (x, y) (x', y')}"
using assms
by auto

text \<open>Compute knight tours of a given length.\<close>

fun knight_tours :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> pos list list" where
  "knight_tours n m 0 = [[]]"
| "knight_tours n m (Suc 0) = map (\<lambda>p. [p]) (List.product [0..<n] [0..<m])"
| "knight_tours n m (Suc l) = concat (map
   (\<lambda>ps. concat (map (\<lambda>p. if p \<in> set ps then [] else [p # ps]) (neighbours n m (hd ps))))
   (knight_tours n m l))"

lemma lengthSucD:
  "length xs = Suc n \<Longrightarrow> \<exists>z zs. xs = z # zs \<and> length zs = n"
by (cases xs) auto

lemma set_knight_tours:
  "set (knight_tours n m l) = {ps. knight_tour n m ps \<and> length ps = l}"
by (induct n m l rule: knight_tours.induct) (force split: prod.splits dest!: lengthSucD)+

(* export_code knight_tours in OCaml *)
(* value "map (length \<circ> knight_tours 3 4) [0..<13]" *)


section \<open>knight tours visiting the whole board\<close>

abbreviation full_knight_tour :: "nat \<Rightarrow> nat \<Rightarrow> pos list \<Rightarrow> bool" where
  "full_knight_tour n m ps \<equiv> knight_tour n m ps \<and> length ps = n * m"

abbreviation full_knight_tours :: "nat \<Rightarrow> nat \<Rightarrow> pos list list" where
  "full_knight_tours n m \<equiv> knight_tours n m (n * m)"

lemma set_full_knight_tours:
  "set (full_knight_tours n m) = {ps. full_knight_tour n m ps}"
using set_knight_tours .

value "full_knight_tours 3 4"
(* value "map (length \<circ> full_knight_tours 3) [0..<9]" *)
(* value "map (length \<circ> full_knight_tours 4) [0..<6]" *)






(* the following material was not shown in the lecture *)






section \<open>Closed knight tours visiting the whole board.\<close>

abbreviation closed_knight_tour :: "nat \<Rightarrow> nat \<Rightarrow> pos list \<Rightarrow> bool" where
  "closed_knight_tour n m ps \<equiv> full_knight_tour n m ps \<and> knight_move (hd ps) (last ps)"

abbreviation closed_knight_tours :: "nat \<Rightarrow> nat \<Rightarrow> pos list list" where
  "closed_knight_tours n m \<equiv> filter (\<lambda>ps. knight_move (hd ps) (last ps)) (knight_tours n m (n * m))"

lemma set_closed_knight_tours:
  "set (closed_knight_tours n m) = {ps. closed_knight_tour n m ps}"
by (auto simp: set_full_knight_tours)

(* value "length (closed_knight_tours 6 6)" *)


section \<open>Detour: parity.\<close>

fun sum_pos :: "pos \<Rightarrow> nat" where
  "sum_pos (x, y) = x + y"

text \<open>Each knight move flips the parity of the sum of the two coordinates\<close>

lemma knight_move_parity:
  "knight_move p q \<Longrightarrow> odd (sum_pos p + sum_pos q)"
by (cases p; cases q) auto

lemma knight_tour_parity:
  assumes "ps \<noteq> []" and "knight_tour n m ps"
  shows "odd (sum_pos (hd ps) + sum_pos (last ps) + length ps)"
proof (cases ps)
  case (Cons p ps)
  show ?thesis using assms(2) unfolding Cons
  proof (induct ps arbitrary: p)
    case (Cons q ps)
    show ?case
      using extend_knight_tour[of _ _ "fst p" "snd p" "q # ps"]
        Cons(1)[of q] Cons(2) knight_move_parity[of p q]
      by auto
  qed auto
qed (simp add: assms(1))

text \<open>Closed tours have even length, hence only exists on boards having even area.\<close>

lemma closed_knight_tour_parity:
  assumes "closed_knight_tour n m ps"
  shows "even (n * m)"
using assms knight_tour_parity[of ps n m] knight_move_parity[of "hd ps" "last ps"]
by (cases ps) auto

section \<open>Heuristic\<close>

text \<open>A liberty is a square that can be reached by a knight move from a given
      square, and has not yet been visited, except by the last move so far.\<close>
definition liberties :: "nat \<Rightarrow> nat \<Rightarrow> pos list \<Rightarrow> pos \<Rightarrow> nat" where
  "liberties n m ps p = length (filter (\<lambda>q. ps = [] \<or> q \<notin> set ps)
    (neighbours n m p))"

definition liberties' :: "nat \<Rightarrow> nat \<Rightarrow> pos list \<Rightarrow> pos \<Rightarrow> nat" where
  "liberties' n m ps p = length (filter (\<lambda>q. ps = [] \<or> q \<notin> set ps \<or> q = hd ps)
    (neighbours n m p))"

text \<open>The heuristic is that at most two unvisited squares can have fewer than 2
  liberties, namely the final square of the tour.\<close>
definition check_liberties :: "nat \<Rightarrow> nat \<Rightarrow> pos list \<Rightarrow> bool" where
  "check_liberties n m ps \<longleftrightarrow> length (filter (\<lambda>p. p \<notin> set ps \<and> liberties n m ps p < 2)
    (List.product [0..<n] [0..<m])) \<le> 2"

fun knight_tours' :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> pos list list" where
  "knight_tours' n m 0 = [[]]"
| "knight_tours' n m (Suc 0) = (map (\<lambda>p. [p]) (List.product [0..<n] [0..<m]))"
| "knight_tours' n m (Suc l) = filter (check_liberties n m) (concat (map
   (\<lambda>ps. concat (map (\<lambda>p. if p \<in> set ps then [] else [p # ps]) (neighbours n m (hd ps))))
   (knight_tours' n m l)))"

abbreviation "full_knight_tours' n m \<equiv> knight_tours' n m (n * m)"

lemma knight_tours'_subset_knight_tours:
  "set (knight_tours' n m l) \<subseteq> set (knight_tours n m l)"
by (induct n m l rule: knight_tours'.induct) auto

lemma full_knight_tour_is_full:
  assumes "full_knight_tour n m ps"
  shows "set ps = {(x, y). x < n \<and> y < m}" (is "_ = ?P")
proof -
  have "card ?P = n * m" by auto
  also have "card (set ps) = n * m" "set ps \<subseteq> ?P"
    using assms by (auto simp: knight_tour_def distinct_card)
  ultimately show ?thesis
    by (intro card_subset_eq) auto
qed

lemma check_partial_knight_tour:
  assumes "full_knight_tour n m (ps @ qs)"
  shows "check_liberties n m qs"
oops

lemma set_knight_tours':
  "{qs |ps qs. full_knight_tour n m (ps @ qs) \<and> length qs = l} \<subseteq> set (knight_tours' n m l)"
oops

lemma set_full_knight_tours':
  "set (full_knight_tours' n m) = {ps. full_knight_tour n m ps}"
oops
(*
using knight_tours'_subset_knight_tours[of n m "n * m"] set_knight_tours'[of n m "n * m"]
by (auto simp: set_knight_tours)
*)

(* value "full_knight_tours' 3 5 = full_knight_tours 3 5" *)
(* value "length (full_knight_tours 3 8)" *)
(* value "length (full_knight_tours' 3 8)" *)
(* value "length (full_knight_tours' 3 9)" *)
(* value "length (full_knight_tours' 3 10)" *)

text \<open>
  ideas for further improvements:
  \<bullet> compute liberties incrementally
  \<bullet> for closed knight tours, one can always start at square (0, 0)
  \<bullet> use better data structures to track visited squares
  \<bullet> restructure search to avoid large intermediate lists
\<close>

section \<open>Just for fun.\<close>

lemma zip_id_tl_Snoc:
  "zip (xs @ [x]) (tl (xs @ [x])) = (if xs = [] then [] else zip xs (tl xs) @ [(last xs, x)])"
proof (induct xs)
  case (Cons x xs) then show ?case by (cases xs) auto
qed auto

lemma zip_id_tl_rev:
  "zip (rev xs) (tl (rev xs)) = map (\<lambda>(x, y). (y, x)) (rev (zip xs (tl xs)))"
proof (induct xs)
  case (Cons x xs) then show ?case
    unfolding rev.simps zip_id_tl_Snoc by (cases xs) auto
qed auto

text \<open>The reverse of a knight tour is a knight tour.\<close>

lemma knight_tour_rev:
  "knight_tour n m (rev ps) \<longleftrightarrow> knight_tour n m ps"
by (auto simp: knight_tour_def zip_id_tl_rev)

lemma closed_knight_tour_cycle1:
  "closed_knight_tour n m (p # ps) \<longleftrightarrow> closed_knight_tour n m (ps @ [p])"
unfolding knight_tour_def rev.simps zip_id_tl_Snoc
by (cases ps) (auto simp: ac_simps)

text \<open>A cyclic shift of a closed knight tour is again a closed knight tour.\<close>

lemma closed_knight_tour_cycle:
  "closed_knight_tour n m (ps @ qs) \<longleftrightarrow> closed_knight_tour n m (qs @ ps)"
proof (induct ps arbitrary: qs)
  case (Cons p ps)
  show ?case
    using Cons[of "qs @ [p]"] closed_knight_tour_cycle1[of n m p "ps @ qs"]
    by (auto simp del: extend_knight_tour)
qed simp

end
