(* Authors:  Christian Sternagel and Bertram Felgenhauer *)

text \<open>
  Summary of induction proofs from last lecture.
  This time in Isabelle/HOL.
\<close>

theory Summary
imports
  Main
  (* "~~/src/HOL/Library/Code_Binary_Nat" *)
  "~~/src/HOL/Library/Code_Target_Nat"
begin

section \<open>Ignore this\<close>

attribute_setup atomized =
  \<open>Scan.succeed (Thm.rule_attribute []
    (Conv.fconv_rule o Object_Logic.atomize o Context.proof_of))\<close>
  "result put into canonical atomic format"

declare [[names_short]]


section \<open>First steps in Isabelle\<close>

subsection \<open>Terms\<close>

term "()"
term "((), ())"

term "1 + 3 :: nat"

value "1 + 3 :: nat"

subsection \<open>Natural deduction\<close>

lemma
  "((P \<longrightarrow> Q) \<longrightarrow> P) \<longrightarrow> P"
proof -
  (* "box proof" - don't use this style in Isabelle! *)
  { assume 1: "(P \<longrightarrow> Q) \<longrightarrow> P"
    { assume 2: "\<not> P"
      { assume 3: "P"
        have 4: "False"                by (rule notE[OF 2 3])
        have 5: "Q"                    by (rule FalseE[OF 4])
      } note "3-5" = this
      have 6: "P \<longrightarrow> Q"                by (rule impI[OF "3-5"])
      have 7: "P"                      by (rule impE[OF 1 6])
      have 8: "False"                  by (rule notE[OF 2 7])
    } note "2-8" = this
    have 9: "P"                        by (rule ccontr [OF "2-8"])
  } note "1-9" = this
  show 10: "((P \<longrightarrow> Q) \<longrightarrow> P) \<longrightarrow> P"   by (rule impI[OF "1-9"])
qed

lemma
  "((P \<longrightarrow> Q) \<longrightarrow> P) \<longrightarrow> P"
proof (rule impI)
  assume 1: "(P \<longrightarrow> Q) \<longrightarrow> P"
  show "P"
  proof (rule ccontr)
    assume 2: "\<not> P"
    have 3: "P \<longrightarrow> Q"
    proof (rule impI)
      assume 4: "P"
      show "Q"
        using 2 4 by (rule notE)
    qed
    have 5: "P"
      using 1 3 by (rule impE)
    show "False"
      using 2 5 by (rule notE)
  qed
qed

lemma peirce:
  "((P \<longrightarrow> Q) \<longrightarrow> P) \<longrightarrow> P"
by auto

lemma beer_drinkers_paradox:
  "\<exists>x. P x \<longrightarrow> (\<forall>y. P y)"
by auto


section \<open>Induction\<close>

subsection \<open>Natural numbers (mathematical induction)\<close>

thm nat.induct
thm nat.induct[atomized]

lemma example1:
  fixes n :: nat
  shows "\<Sum>{0..n} = n * (n + 1) div 2"
proof (induct n)
  case 0 show ?case by simp
next
  case (Suc n) then show ?case by simp
qed

subsection \<open>Lists (structural induction)\<close>

(* type a list = Nil | Cons of (a * a list) *)

datatype 'a list = Nil | Cons 'a "'a list"

thm list.induct
thm list.induct[atomized]

fun append :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where
  "append Nil ys = ys"
| "append (Cons x xs) ys = Cons x (append xs ys)"

lemma example2:
  "append xs Nil = xs"
by (induct xs) simp_all

fun length :: "'a list \<Rightarrow> nat"
where
  "length Nil = 0"
| "length (Cons x xs) = Suc (length xs)"

lemma example3:
  "length (append xs ys) = length xs + length ys"
by (induct xs) auto

subsection \<open>Binary trees\<close>

datatype 'a btree = Empty |  Node "'a btree" 'a "'a btree"

thm btree.induct

fun height :: "'a btree \<Rightarrow> nat"
where
  "height Empty = 0"
| "height (Node l x r) = max (height l) (height r) + 1"

fun size :: "'a btree \<Rightarrow> nat"
where
  "size Empty = 0"
| "size (Node l x r) = size l + size r + 1"

fun perfect :: "'a btree \<Rightarrow> bool"
where
  "perfect Empty \<longleftrightarrow> True"
| "perfect (Node l x r) \<longleftrightarrow> height l = height r \<and> perfect l \<and> perfect r"

lemma example4:
  "perfect t \<longrightarrow> size t = (2 ^ height t) - 1"
by (induct t) (auto simp: diff_Suc split: nat.splits)

section \<open>Code Generation\<close>

export_code open
  append (* length height *)
  in OCaml (*Haskell, SML (Standard ML), Scala*)
  module_name Test

end
